import warnings
import time

import spidev

from max7219simple import font

MAX7219_REG_NOOP        = 0x0
MAX7219_REG_DIGIT0      = 0x1
MAX7219_REG_DIGIT1      = 0x2
MAX7219_REG_DIGIT2      = 0x3
MAX7219_REG_DIGIT3      = 0x4
MAX7219_REG_DIGIT4      = 0x5
MAX7219_REG_DIGIT5      = 0x6
MAX7219_REG_DIGIT6      = 0x7
MAX7219_REG_DIGIT7      = 0x8
MAX7219_REG_DECODEMODE  = 0x9
MAX7219_REG_INTENSITY   = 0xA
MAX7219_REG_SCANLIMIT   = 0xB
MAX7219_REG_SHUTDOWN    = 0xC
MAX7219_REG_DISPLAYTEST = 0xF

class Strip:
    def __init__(self, size, spi_bus=0, spi_device=0, rotate=False):
        self.size = size
        self._rotate = rotate

        self._font = font.DEFAULT_FONT

        self._spi = spidev.SpiDev()
        self._spi.open(spi_bus, spi_device)

        self._buffer = bytearray(b'\0' * 8 * size)

        self.command(MAX7219_REG_SCANLIMIT,   7) # show all 8 digits
        self.command(MAX7219_REG_DECODEMODE,  0) # use matrix (not digits)
        self.command(MAX7219_REG_DISPLAYTEST, 0) # no display test
        self.command(MAX7219_REG_SHUTDOWN,    1) # not shutdown mode

        self.brightness(7)                        # intensity: range: 0..15
        self.commit()

    def command(self, register, data):
        assert MAX7219_REG_DECODEMODE <= register <= MAX7219_REG_DISPLAYTEST
        self._write([register, data] * self.size)

    def _write(self, data):
        self._spi.xfer2(list(data))

    def brightness(self, intensity):
        assert 0 <= intensity < 16, "Invalid brightness: {0}".format(intensity)
        self.command(MAX7219_REG_INTENSITY, intensity)

    def clear(self):
        for i in range(len(self._buffer)):
            self._buffer[i] = 0

    def commit(self):
        for row in range(8):
            message = bytearray()

            reg  = MAX7219_REG_DIGIT0 + row
            data = self._buffer[row::8]

            for matrix in range(self.size):
                message.append(reg)
                message.append(data[matrix])
            self._write(message)

    def __getitem__(self, coordinates):
        assert len(coordinates) == 2
        row, col = coordinates
        assert 0 <= row < 8

        if self._rotate:
            row = 7 - row
            col = 8 * self.size - 1 - col

        row = row + col//8*8
        col = col % 8

        buffer = self._buffer

        return (buffer[row] >> col) & 1;

    def __setitem__(self, coordinates, value):
        assert len(coordinates) == 2
        row, col = coordinates
        assert 0 <= row < 8

        if self._rotate:
            row = 7 - row
            col = 8 * self.size - 1 - col

        row = row + col//8*8
        col = col % 8

        buffer = self._buffer

        buffer[row]  ^= (-value ^ buffer[row]) & (1 << col)

    def letter(self, column, code, proportional=True):
        try:
            character = self._font[code]
        except IndexError:
            warnings.warn("Unknow code in the selected font: %i" % code)
            character = self._font[ord("_")]

        if proportional:
            character = [ col for col in character if col != 0x00 ]

        if code == 0x20:
            width = 8
        else:
            width = len(character) + 1

        min_column = max(0, -column)
        max_column = min(len(character), self.size * 8 - column)

        for col in range(min_column, max_column):
            for line in range(8):
                bit = character[col] >> line & 1
                self[line, column + col] = bit

        return width

    def message(self, message, firstcol=0):
        bytes = message.encode("CP437")
        col = firstcol
        for code in bytes:
            col += self.letter(col, code)
        return col

    def scroll(self, message, start=None, stop=None, delay=0.01, loop=-1, direction=-1):
        if direction > 0:
            direction = +1

            if start is None:
                start = - 8 * len(message)
            if stop is None:
                stop = self.size * 8
        else:
            direction = -1

            if start is None:
                start = self.size * 8
            if stop is None:
                stop = - 8 * len(message)

        count = 0
        while True:
            for col in range(start, stop, direction):
                self.clear()
                self.message(message, col)
                self.commit()
                time.sleep(delay)

            count += 1
            if loop > 0 and count >= loop:
                break

    def fill(self):
        for line in range(8):
            for col in range(8 * self.size):
                self[line, col] = 1
